/*
 * ani-parser/checkifeval.cc
 * 
 * Tree evaluation check. Well, it does not do a lot but is necessary. 
 * This is part of the AniVision project. 
 * 
 * Copyright (c) 2003 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#include "tree.h"
#include "coreglue.h"
#include "opfunc.h"

#include <stdio.h>

// NOTE: 
// These routines do basically the same as the CheckIfEvaluable() and Exec() routines 
// just that they check if we actually CAN eval, i.e. the eval funcs 
// are set etc. 
// MAKE SURE TO KEEP THEM IN SYNC WITH CheckIfEvaluable() AND Exec(). 

namespace ANI
{

int TreeNode::CheckIfEvaluable()
{
	// Must be overridden if called. 
	assert(0);
}

int TNStatement::CheckIfEvaluable()
{
	// Must be overridden by all TNStatement-derived classes. 
	assert(0);
}

/*----------------------------------------------------------------------------*/

int TNIdentifier::CheckIfEvaluable()
{
	int fail=0;
	if(!evaladrfunc)
	{
		Error(GetLocationRange(),
			"invalid use of \"%s\"\n",
			CompleteStr().str());
		++fail;
	}
	return(fail);
}


int TNExpression::CheckIfEvaluable()
{
	// Special case as usual :)
	// I mean: TNExpression is a base class for different expressions 
	// the only special case of no derived class is an expression 
	// holding an identifier. 
	if(!down.first() || down.first()->NType()!=TN_Identifier || 
		down.first()!=down.last())
	{  assert(0);  }
	
	return(down.first()->CheckIfEvaluable());
}


int TNExpressionList::CheckIfEvaluable()
{
	int fail=0;
	for(TreeNode *i=down.first(); i; i=i->next)
	{
		fail+=i->CheckIfEvaluable();
	}
	return(fail);
}


int TNValue::CheckIfEvaluable()
{
	return(0);
}


int TNArray::CheckIfEvaluable()
{
	assert(assfunc);
	
	int fail=0;
	for(TreeNode *i=down.first(); i; i=i->next)
	{
		fail+=i->CheckIfEvaluable();
	}
	return(fail);
}


int TNFCallArgument::CheckIfEvaluable()
{
	// This is simple as the identifer mapping was already done. 
	// Simply evaluate the expresion: 
	return( down.last()->CheckIfEvaluable() );
}


int TNAssignment::CheckIfEvaluable()
{
	int fail=0;
	
	// If this assert fails and it is due to user error 
	// which did not get caught yet, replace it by an error 
	// message like the one in TNIdentifier. 
	assert(assfunc);
	
	fail+=down.last()->CheckIfEvaluable();
	fail+=down.first()->CheckIfEvaluable();
	
	return(fail);
}


int TNOperatorFunction::CheckIfEvaluable()
{
	int fail=0;
	
	switch(func_id)
	{
		case OF_PODCast:
			fail+=down.last()->CheckIfEvaluable();
			break;
		case OF_MembSel:
		case OF_Mapping:
			fail+=down.first()->CheckIfEvaluable();
			// last() is only evaluated if we're not POD-indexing. 
			if(!ms_is_subscript)
			{  fail+=down.last()->CheckIfEvaluable();  }
			break;
		case OF_NewArray:
			fail+=down.first()->next->CheckIfEvaluable();
			break;
		case OF_Delete:
			// Not yet implemented. 
			Error(GetLocationRange(),
				"operator delete not yet implemented\n");
			abort();
			break;
		case OF_PostIncrement:  // fall through.......
		case OF_PostDecrement:
		case OF_PreIncrement:
		case OF_PreDecrement:
		case OF_FCall:
		case OF_Subscript:
		case OF_New:
		default:
			for(TreeNode *tn=down.first(); tn; tn=tn->next)
			{  fail+=tn->CheckIfEvaluable();  }
			break;
	}
	
	// If this assert fails and it is due to user error 
	// which did not get caught yet, replace it by an error 
	// message like the one in TNIdentifier. 
	assert(evalfunc);
	
	return(fail);
}


int TNDeclarator::CheckIfEvaluable()
{
	int fail=0;
	
	// If this assert fails and it is due to user error 
	// which did not get caught yet, replace it by an error 
	// message like the one in TNIdentifier. 
	assert(evalfunc);
	
	// This is non-recursively, so this check tests if the 
	// decl has an initializer or not: 
	if(decltype==DT_Initialize)
	{
		assert(assfunc);  // Only needed if DT_Initialize. 
		fail+=down.last()->CheckIfEvaluable();
	}
	
	// Check the identifier (find DT_Name): 
	TreeNode *tn=this;
	for(; tn->NType()==TN_Declarator; tn=tn->down.first());
	TNIdentifier *idf=(TNIdentifier*)tn;
	assert(idf->NType()==TN_Identifier);
	fail+=idf->CheckIfEvaluable();
	
	return(fail);
}


int TNNewArraySpecifier::CheckIfEvaluable()
{
	int fail=0;
	
	assert(!down.is_empty());
	assert(evalfunc);
	
	// This is special because we walk left-to-right instead 
	// of top-to-bottom. 
	if(next && !next->down.is_empty())
	{  fail+=next->CheckIfEvaluable();  }
	
	return(fail);
}


int TNExpressionStmt::CheckIfEvaluable()
{
	int fail=0;
	if(down.first())
	{
		fail+=down.first()->CheckIfEvaluable();
	}
	return(fail);
}


int TNIterationStmt::CheckIfEvaluable()
{
	int fail=0;
	
	if(for_init_stmt)
	{  fail+=for_init_stmt->CheckIfEvaluable();  }
	if(loop_cond)
	{  fail+=loop_cond->CheckIfEvaluable();  }
	assert(loop_stmt);
	{  fail+=loop_stmt->CheckIfEvaluable();  }
	if(for_inc_stmt)
	{  fail+=for_inc_stmt->CheckIfEvaluable();  }
	
	return(fail);
}


int TNSelectionStmt::CheckIfEvaluable()
{
	int fail=0;
	
	assert(cond_expr);
	assert(if_stmt);
	
	fail+=cond_expr->CheckIfEvaluable();
	fail+=if_stmt->CheckIfEvaluable();
	if(else_stmt)
	{  fail+=else_stmt->CheckIfEvaluable();  }
	
	return(fail);
}


int TNJumpStmt::CheckIfEvaluable()
{
	int fail=0;
	
	switch(jumptype)
	{
		case JT_Break:
			break;
		case JT_Return:
			assert(evalfunc);
			if(down.first())
			{  fail+=down.first()->CheckIfEvaluable();  }
			break;
		case JT_None:  // fall through
		default: assert(0);
	}
	
	return(fail);
}


int TNCompoundStmt::CheckIfEvaluable()
{
	int fail=0;
	
	for(TreeNode *tn=down.first(); tn; tn=tn->next)
	{
		fail+=tn->CheckIfEvaluable();
	}
	
	return(fail);
}


int TNDeclarationStmt::CheckIfEvaluable()
{
	int fail=0;
	
	// Process all declarators: 
	for(TreeNode *tn=down.first()->next; tn; tn=tn->next)
	{
		fail+=tn->CheckIfEvaluable();
	}
	
	return(fail);
}


int TNAniDescStmt::CheckIfEvaluable()
{
	int fail=0;
	for(TreeNode *tn=down.first(); tn; tn=tn->next)
	{  fail+=tn->CheckIfEvaluable();  }
	return(fail);
}


int TNFuncDefArgDecl::CheckIfEvaluable()
{
	// Check the delarator: 
	return( down.last()->CheckIfEvaluable() );
}


//------------------------------------------------------------------------------

int TNFunctionDef::CheckIfEvaluable()
{
	int fail=0;
	
	//fprintf(stderr,"CHECKING: function %s\n",
	//	((TNIdentifier*)down.first())->CompleteStr().str());
	
	//fail+=func_name->CheckIfEvaluable();  NO! (No need for eval; this is DEF not CALL.)
	for(TreeNode *tn=first_arg; tn; tn=tn->next)
	{  fail+=tn->CheckIfEvaluable();  }
	fail+=func_body->CheckIfEvaluable();
	
	return(fail);
}


int TNObject::CheckIfEvaluable()
{
	int fail=0;
	for(TreeNode *tn=down.first()->next; tn; tn=tn->next)
	{  fail+=tn->CheckIfEvaluable();  }
	return(fail);
}


int TNSetting::CheckIfEvaluable()
{
	int fail=0;
	for(TreeNode *tn=down.first()->next; tn; tn=tn->next)
	{  fail+=tn->CheckIfEvaluable();  }
	return(fail);
}


int TNAnimation::CheckIfEvaluable()
{
	int fail=0;
	for(TreeNode *tn=down.first(); tn; tn=tn->next)
	{  fail+=tn->CheckIfEvaluable();  }
	return(fail);
}

}  // end of namespace func
