/*
 * calccore/ccif.h
 * 
 * Calc core interface to animation core. 
 * This is part of the AniVision project. 
 * 
 * Copyright (c) 2003 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#ifndef _ANIVISION_CALCCORE_CCIF_H_
#define _ANIVISION_CALCCORE_CCIF_H_

#include <hlib/cplusplus.h>
#include <valtype/valtypes.h>

// -*- MAIN CALCULAION CORE INTERFACE SPECIFICATION -*-

namespace NUM
{
	// See omatrix.h: 
	class ObjectMatrix;
}
class ExecThread;


namespace CC   // calculation core namespace
{

// See "adtree.h": 
class ADTNAniDescBlock;


// NOTE: All the types in here are NOT C++-save. Always use 
//       pointers to them. 

// Internal bug abort routine; pass line number as argument.  
extern void _ccif_failassert(int line);


// NOTE: To create a CCIF_* object, you need to ask the 
//       CCIF_Factory. This allows for different core 
//       implementations. 
class CCIF_Factory;

// Calculation Core InterFace class for SETTING:
// Meant to be attached to a setting. Just a base for 
// the actual class provided by some deeper implementetion. 
class CCIF_Setting
{
	friend class CCIF_Factory;
	protected:
		// Current setting time. 
		// This is set by ComputeAll() and SetTimeRange(). 
		double current_time;
		
		CCIF_Setting();  // Use factory to create. 
	private:
		CCIF_Setting &operator=(const CCIF_Setting &)
			{  _ccif_failassert(__LINE__);  return(*this);  }
		CCIF_Setting(const CCIF_Setting &)
			{  _ccif_failassert(__LINE__);  }
	public:  _CPP_OPERATORS
		virtual ~CCIF_Setting();
		
		// Get current setting time: 
		double CurrentSettingTime() const
			{  return(current_time);  }
		
		// Set the time range for the animation of this setting. 
		// Return value: 
		//  0 -> OK
		virtual int SetTimeRange(const Range &tr);
		
		// Compute all values and objects for the current time. 
		virtual int ComputeAll();
		
		// Advance time. Returns 1 if the animation now ends, 
		// 0 if to continue. 
		virtual int TimeStep();
};


// Calculation Core InterFace class for OBJECT: 
// This class is meant to be attached to an object to represent 
// the object's orientation in space and its time. 
// Just a base for the actual class provided by some deeper 
// implementetion. 
class CCIF_Object
{
	friend class CCIF_Factory;
	public:
		// Value flags: 
		enum
		{
			VF_None=    0x00,  // no flag
			VF_Time=    0x01,  // time
			VF_Location=0x02,  // pos, front, up, speed, accel
			VF_Active=  0x04,  // active flag (integer, not bool)
			VF_TimeLocAct=VF_Time|VF_Location|VF_Active,
		};
		
		// Provided variables for the object, depending on the 
		// variable flags. 
		enum OVal  // "object value"
		{
			ValNone=0,     // void
			// VF_Location
			ValPos,        // 3d vector
			ValFront,      // 3d vector
			ValUp,         // 3d vector
			ValSpeed,      // 3d vector
			ValAccel,      // 3d vector
			ValMatrix,     // position and orientation as matrix -> ObjectMatrix
			// VF_Time: 
			ValTime,       // double
			// VF_Active: 
			ValActive,     // int
		};
		
	protected:
		
		// Associated setting interface: 
		CCIF_Setting *setting;
		// Present VF_* flags: 
		int vflags;
		
		CCIF_Object(CCIF_Setting *_setting,int _vflags);  // Use factory to create. 
	private:
		CCIF_Object &operator=(const CCIF_Object &)
			{  _ccif_failassert(__LINE__);  return(*this);  }
		CCIF_Object(const CCIF_Object &)
			{  _ccif_failassert(__LINE__);  }
	public:  _CPP_OPERATORS
		virtual ~CCIF_Object();
		
		// Check if the passed variable is provided (checks against flags): 
		// Return value: 
		//  1 -> yes
		//  0 -> no
		int ProvidesVal(OVal ov) const;
		
		// Get current value of passed variable. 
		// Stores value under passed pointer. ObjectMatrix
		// Vector values expect to find an array double[3]
		// while scalar values only need pointer to double. 
		// Return value: 
		//  0 -> OK
		//  1 -> no such value (-> ProvidesVal())
		//  2 -> incompatible type (e.g. integer version 
		//       for ValPos...)
		virtual int GetVal(OVal ov,int *store_here);
		virtual int GetVal(OVal ov,double *store_here);
		virtual int GetVal(OVal ov,NUM::ObjectMatrix *store_here);
		
		// This is used "internally" if an ani desc block for the 
		// associated object is evaluated (i.e. "executed"). 
		virtual int EvalAniDescBlock(ADTNAniDescBlock *adb,ExecThread *et);
};


// Calculation Core InterFace class for VARIABLE:
// Animate a value; available for n-dim vector (n may be 1). 
// Just a base for the actual class provided by some deeper 
// implementetion. 
class CCIF_Value
{
	friend class CCIF_Factory;
	protected:
		// Associated setting interface: 
		CCIF_Setting *setting;
		// Dimension of the var vector. 
		int dimension;
		
		CCIF_Value(CCIF_Setting *_setting,int _dimension);  // Use factory to create. 
	private:
		CCIF_Value &operator=(const CCIF_Value &)
			{  _ccif_failassert(__LINE__);  return(*this);  }
		CCIF_Value(const CCIF_Value &)
			{  _ccif_failassert(__LINE__);  }
	public:  _CPP_OPERATORS
		virtual ~CCIF_Value();
};


// The core factory. This is the place where you choose the core 
// to be used. First, create the factory object of the core you 
// like. Then use the factory to create all the other 
// CCIF_* types. 
// This is just the base class for the different factories. 
class CCIF_Factory
{
	protected:
		CCIF_Factory(const char *name);  // Must create a derived class. 
	private:
		const char *name;   // non-alloc
		
		CCIF_Factory &operator=(const CCIF_Factory &)
			{  _ccif_failassert(__LINE__);  return(*this);  }
		CCIF_Factory(const CCIF_Factory &)
			{  _ccif_failassert(__LINE__);  }
	public:  _CPP_OPERATORS
		virtual ~CCIF_Factory();
		
		// Get the name of the core: 
		const char *Name() const
			{  return(name);  }
		
		// Create a setting interface: (Use delete to destroy again.)
		virtual CCIF_Setting *CreateCCIF_Setting();
		
		// Create an object interface: For the vflags, see 
		// the enum VF_* above. 
		// You need to pass the associated setting interface as 
		// first arg. 
		// Note that VF_Location and VF_Active require VF_Time. 
		// (Use delete to destroy again.)
		virtual CCIF_Object *CreateCCIF_Object(CCIF_Setting *setting,
			int vflags);
		
		// Create a value interface: (Use delete to destroy again.)
		virtual CCIF_Value *CreateCCIF_Value(CCIF_Setting *setting,
			int dimension);
};

}  // end of namespace CC

#endif  /* _ANIVISION_CALCCORE_CCIF_H_ */
