/*
 * core/a_setting.cc
 * 
 * ANI tree setting scope. 
 * This is part of the AniVision project. 
 * 
 * Copyright (c) 2003 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#include "animation.h"
#include <ani-parser/treereg.h>

#include "execthread.h"
#include <calccore/ccif.h>


using namespace ANI;


int AniSetting::EvalFunc_FCallSetting::Eval(ExprValue &retval,
	TNOperatorFunction *tn,ExecThreadInfo *info)
{
	EvalStateStack *ess=info ? info->ess : NULL;
	
	ExprValue t_val;
	switch(ess->popS())
	{
		case 0:  // arg evaluation
		{
			ExprValue tmp;
			// Eval the one and only arg: 
			if(tn->down.first()->next->Eval(tmp,info))
			{
				ess->pushS(0);
				return(1);
			}
			assert(arg_assfunc_t);
			arg_assfunc_t->Assign(/*lvalue=*/t_val,/*rvalue=*/tmp);
		} goto skippop;
		case 1:
			ess->popEV(&t_val);
		skippop:;
			// We could allow a return value here (but then must 
			// also change the function code to set ret_type to 
			// integer/scalar for the TF step. 
			if(setting->_DO_EvalSetting(t_val,info))
			{
				ess->pushEV(t_val);
				ess->pushS(1);
				return(1);
			}
			retval=ExprValue::CVoidType;
			break;
		default: assert(0);
	}
	return(0);	
}

AniSetting::EvalFunc_FCallSetting::~EvalFunc_FCallSetting()
{
	setting=NULL;
	DELETE(arg_assfunc_t);
}

//------------------------------------------------------------------------------


static ExprValueType _RangeType()
{
	Value::CompleteType tmp;
	tmp.type=Value::VTRange;
	return(ExprValueType(tmp));
}

int AniSetting::CG_LookupFunction(AniGlue_ScopeBase::LookupFunctionInfo *lfi)
{
	// Inside the animation block, one can call settings 
	// like functions. 
	// We need opfunc pointer for the checks. 
	assert(lfi->opfunc);
	// See if this is inside an animation block: 
	// Find function def. 
	TreeNode *tn=lfi->opfunc->parent();
	for(; tn && tn->NType()!=TN_FunctionDef; tn=tn->parent());
	if(!tn || !((TNFunctionDef*)tn)->function || !(
		((AniFunction*)(((TNFunctionDef*)tn)->function))->GetAttrib() & 
			AniFunction::IS_Animation) )
	{
		Error(lfi->opfunc->down.first()->GetLocationRange(),
			"cannot call setting %s like a function outside animation block\n",
			CompleteName().str());
		return(0);
	}
	
	// So, we're inside an animation block and the user calls 
	// a setting to be processed. 
	// That looks like this: 
	// "setting(t=<range>,time=func_of($t));"
	
	// This especially means that the setting is given directly 
	// as an identifier. Silly jokes such as 
	//   (val ? TheSetting : TheSetting)(t=1..2); 
	// will not work (there is no point in doing the above). 
	// Hence, we check here if there is an identifier as the first 
	// child if the function call. If so, assign an eval func to 
	// it. If not, simply go on without doing so, we'll catch an 
	// "inalid use of ..." error lateron in checkifeval.cc. 
	do {
		TreeNode *nnode=lfi->opfunc->down.first();
		if(nnode->NType()!=TN_Expression) break;
		if(((TNExpression*)nnode)->ExprType()!=TNE_Expression) break;
		if(nnode->down.first()->NType()!=TN_Identifier) break;
		TNIdentifier *idf=(TNIdentifier*)nnode->down.first();
		assert(!idf->evaladrfunc);
		// This eval func may do just anything. It will never be 
		// called. 
		idf->evaladrfunc=new AniUserVariable::EAF_NULL();
	} while(0);
	
	// Do some simple argument checking. Need a better version once 
	// there are more args. 
	// Time trafo (time=f($t)) not yet supported. 
	
	// Argument index: 
	int time_idx=-1;
	
	// Create eval function handler: 
	EvalFunc_FCallSetting *evalfunc=new EvalFunc_FCallSetting(this);
	
	bool match_ok=0;
	do {
		
		char which_arg='\0'; 
		for(int i=0; i<lfi->n_args; i++)
		{
			if(lfi->arg_name[i])
			{
				// Look up argument: 
				RefString name=lfi->arg_name[i]->CompleteStr();
				if(name=="time")  which_arg='t';
				else goto breakmatch;
			}
			else switch(which_arg)  // <-- switch forward one arg
			{
				case '\0':  which_arg='a';  break;
				case 't':   goto breakmatch;
				default:  assert(0);
			}
			
			// Check arg type and repetition:
			int rv=-10;
			switch(which_arg)
			{
				case '\0':  goto breakmatch;
				case 't':
					if(time_idx>=0)  goto breakmatch;
					time_idx=i;
					rv=_RangeType().CanAssignType(lfi->arg_type[i],
						&evalfunc->arg_assfunc_t);
					break;
				default:  assert(0);
			}
			if(rv==0)  // No, cannot. Type mismatch. 
			{  goto breakmatch;  }
		}
		
		// Currently, time arg MUST be present. 
		if(time_idx<0) break;
		
		// Looks good. 
		match_ok=1;
	} while(0); breakmatch:;
	if(!match_ok)
	{
		Error(lfi->opfunc->GetLocationRange(),
			"no matching call to setting %s\n",
			FuncCallWithArgsString(CompleteName(),lfi).str());
		DELETE(evalfunc);
		return(0);
	}
	
	// Okay, store eval info: 
	lfi->ani_function=this;
	lfi->ret_evt=ExprValueType::CVoidType;
	lfi->evalfunc=evalfunc;
	lfi->may_try_ieval=0;
	
	return(0);
}


int AniSetting::_DO_EvalSetting(const ExprValue &t_val,ExecThreadInfo *info)
{
	Value tmp;
	t_val.GetPOD(&tmp);
	const Range *timerange=tmp.GetPtr<Range>();
	
	// Now we "eval" the setting. That means it "was called like 
	// a function". And that actually means: we run the animation 
	// in this setting. 
	
	// Context name. Currently, this is always the default name "". 
	String ctx_name("");
	
	// A verbose message: 
	fprintf(stderr,"Running setting %s for time %s using context \"%s\".\n",
		CompleteName().str(),
		timerange->ToString().str(),
		ctx_name.str());
	
	// TODO: Check name; create new context if needed; make the 
	// context the active one; create new exec thread if needed. 
	// Let's do it: 
	AniAnimation *ani=(AniAnimation*)Parent();
	assert(ani->ASType()==AST_Animation);
	// This will return the context with the passed name or 
	// allocate a new one if it did not yet exist. 
	AniAnimation::Context *ctx=ani->GetContextByName(ctx_name,this);
	
	// Set the time range: 
	ctx->ccif->SetTimeRange(*timerange);
	
	// Actually switch contexts and run them: 
	ani->RunContext(ctx);
	
	return(0);
}


AniSetting::AniSetting(const RefString &name,ANI::TreeNode *_treenode,
	AniAniSetObj *_parent) : 
	AniAniSetObj(AST_Setting,name,_treenode,_parent)
{
	settinginit_eval_func=NULL;
	
	// Register internal functions (atan(),sin(),$pov(),...) 
	// and variables ($time,...): 
	AniInternalFunction::RegisterInternalFunctions(this);
	AniInternalVariable::RegisterInternalVariables(this);
}

AniSetting::~AniSetting()
{
	DELETE(settinginit_eval_func);
}

