/*
 * core/aniinstance.h
 * 
 * Scope (and object in particular) instance. 
 * This is part of the AniVision project. 
 * 
 * Copyright (c) 2003 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#ifndef _ANIVISION_CORE_ANIINSTANCE_H_
#define _ANIVISION_CORE_ANIINSTANCE_H_ 1

#include <core/aniscope.h>
#include <core/a_variable.h>
//#include <core/animation.h>

// For attached POV objects: 
#include <pov-core/povparser.h>

// Include the calculation core interface: 
//#include <calccore/ccif.h>
namespace CC
{
	class CCIF_Setting;
	class CCIF_Object;
	class CCIF_Value;
}

// Non-instances (i.e. (proto)types): 
class AniAniSetObj;
class AniAnimation;
class AniSetting;
class AniObject;

class Animation_Context;

// This is the base class for ani scope instances, i.e. 
// animation, setting, object, function, anonymous scope. 
// All the associated (member/...) variables are stored in 
// here and accessed O(1) via array subscript. 
// See below for derived types. 
// NOTE: 
//  - This is an AniObjectInstance if the attached AniScopeBase *asb 
//    refers to an object. 
//  - This is a plain AniScopeInstanceBase otherwise. 
class AniScopeInstanceBase : public ANI::AniGlue_ScopeInstanceBase
{
	public:
		static AniScopeInstanceBase *CreateAniInstance(AniScopeBase *asb);
		
		typedef AniInternalVariable::BuiltinVarType BuiltinVarType;
	protected:
		// Pointer to associated AniScopeBase: 
		AniScopeBase *asb;
		
		// Non-static, non-automatic regular member variables 
		// (i.e. all which are registered in the ani scope tree). 
		// Array of size n_memb_vars. 
		int n_memb_vars;
		ANI::ExprValue *memb_vars;
		
		// NEVER CALL DIRECTLY. USE CreateAniInstance(). 
		// Constructor. Will set asb an n_memb_vars and allocate 
		// memb_vars[]. Setting up memb vars is task of derived 
		// class. 
		AniScopeInstanceBase(AniScopeBase *asb);
	public:  _CPP_OPERATORS
		// Destructor: free memb_vars[]. 
		~AniScopeInstanceBase();
		
		// Get type... (of attached ASB): 
		AniScopeBase::AniScopeType ASType() const
			{  return(asb->ASType());  }
		// ...and ASB: 
		AniScopeBase *GetASB() const
			{  return(asb);  }
		
		// Get member variable associated with passed 
		// addressation index. O(1) and lvalue-capable: 
		// NOTE: Returns a true C++ reference. 
		ANI::ExprValue &GetMemberVar(int idx) const;
		
		// Get internal variable: 
		// Calls assign() on passed ExprValue. 
		// Return value: 
		//    0 -> OK
		//  1,2 -> see CCIF_Object::GetVal()
		//   -2 -> ccif=NULL: no interface
		// In case of error, object instances store 0 in the passed var 
		// other instances abort immediately. 
		virtual int GetInternalVar(BuiltinVarType bvt,
			ANI::ExprValue &store_here);
		
		// [overriding virtuals from AniGlue_ScopeInstanceBase:] 
		String ToString() const;
		void GetExprValueType(ANI::ExprValueType *store_here) const;
};


class AniObjectInstance : public AniScopeInstanceBase
{
	public:
		// There may be classes which need to know when a certain 
		// object instance is deleted (notably the POV attachment 
		// code). For EACH AniObjectInstance you monitor, create 
		// ONE DeleteNotifier-derived class and register it using 
		// RegisterDeleteNotifier(). When the object is deleted, 
		// the DeleteNotify() function will be called. 
		// NOTE: THIS FUNCTION MUST THEN CALL UnregisterDeleteNotifier() 
		//       to remove the notifier from the list. 
		// NOTE: You need ONE DeleteNotifier for EACH AniObjectInstance 
		//       but an AniObjectInstance can have any number of 
		//       DeleteNotifiers registered. 
		//** The DeleteNotifier is now externally defined as a template. 
		
	private:
		// The list of attached POV objects is (now) in the 
		// animation context. 
		
		// This is a list of attached delete notifiers. 
		// See above for more info. 
		LinkedList<DeleteNotifier<AniObjectInstance> > del_notifier_list;
		
		// Calculatio Core InterFace for Objects: 
		CC::CCIF_Object *ccif;
		
		// Get AniAnimation pointer: 
		AniAnimation *_GetAnimation();
	public:
		AniObjectInstance(AniScopeBase *asb);
		~AniObjectInstance();
		
		// Internally used by AniAnimation for POV object attach/detach. 
		ANI::TNCompoundStmt *_GetPOVCompound();
		
		// Get calc core interface: 
		CC::CCIF_Object *GetCCIF() const
			{  return(ccif);  }
		
		// [overriding virtual from AniScopeInstanceBase:]
		int GetInternalVar(BuiltinVarType bvt,ANI::ExprValue &store_here);
		
		// See above...
		void RegisterDeleteNotifier(DeleteNotifier<AniObjectInstance> *dn)
			{  del_notifier_list.append(dn);  }
		void UnregisterDeleteNotifier(DeleteNotifier<AniObjectInstance> *dn)
			{  del_notifier_list.dequeue(dn);  }
};

#endif  /* _ANIVISION_CORE_ANIINSTANCE_H_ */
