/*
 * numerics/spline/lspline.cc
 * 
 * Numerics library linear spline. 
 * 
 * Copyright (c) 2003 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#include "lspline.h"

namespace NUM
{

const SplineBase::Properties *LinearSpline::SplineProperties() const
{
	static const Properties prop=
	{
		splinetype: ST_LSpline,
		name: "lspline",
		cont_order: 0
	};
	return(&prop);
}


void LinearSpline::Clear()
{
	SplineBase::Clear();
}


int LinearSpline::Eval(double t,double *result) const
{
	if(!n)
	{  for(int j=0; j<dim; j++)  result[j]=0.0;  return(-2);  }
	
	// Compute the index of the polynomial: 
	int retval;
	int idx=_EvalFindIdx(t,&retval);
	
	// Normalize t: 
	t-=get_t(idx);
	
	// Evaluate the spline: 
	double h = tv ? (tv[idx+1]-tv[idx]) : 1.0;
	for(int j=0; j<dim; j++)
	{  result[j] = (xv(idx+1)[j]-xv(idx)[j])*t/h + xv(idx)[j];  }
	
	return(retval);
}

int LinearSpline::EvalD(double t,double *result) const
{
	if(!n)
	{  for(int j=0; j<dim; j++)  result[j]=0.0;  return(-2);  }
	
	// Compute the index of the polynomial: 
	int retval;
	int idx=_EvalFindIdx(t,&retval);
	
	// t not needed below here. 
	
	// "Evaluate" the spline: 
	double h = tv ? (tv[idx+1]-tv[idx]) : 1.0;
	for(int j=0; j<dim; j++)
	{  result[j] = (xv(idx+1)[j]-xv(idx)[j])/h;  }
	
	return(retval);
}

int LinearSpline::EvalDD(double t,double *result) const
{
	// See description. 
	for(int j=0; j<dim; j++)
	{  result[j]=0.0;  }
	
	return(0);
}


int LinearSpline::Create(const VectorArray<double> &cpoints,const double *tvals)
{
	if(cpoints.NVectors()<2 || cpoints.Dim()<(tvals==TVALS_CPTS ? 2 : 1))
	{  return(-2);  }
	
	// Check t values if specified: 
	if(tvals && tvals!=TVALS_DIST && 
		( (tvals==TVALS_CPTS) ? 
			_TestAscentingOrder_LastCompo(cpoints) : 
			_TestAscentingOrder(tvals,cpoints.NVectors()) ) )
	{  return(-4);  }
	
	// Get rid of old stuff...
	// ...and copy the control points and t values, if present: 
	_CopyPointsAndTVals(cpoints,tvals);  // calls Clear(). 
	
	return(0);
}


double LinearSpline::CalcLength(NUM::Integrator_R_R * /*integrator*/,
	int /*int_by_piece*/)
{
	if(!n) return(0.0);
	double len=0.0;
	double *p0,*p1=xv(0);
	for(int i=1; i<=n; i++)
	{
		p0=p1;
		p1=xv(i);
		len+=NUM::DistEuclidic(p0,p1,dim);
	}
	return(len);
}


LinearSpline::LinearSpline() : 
	SplineBase()
{
	// nothing to do
}

LinearSpline::~LinearSpline()
{
	Clear();
}

}  // end of namespace NUM
